#include <stdio.h>
#include <string.h>
#include <math.h>
#include "pico/stdlib.h"
#include "hardware/spi.h"
#include "hardware/adc.h"
#include "./font.h"

#include "pico/multicore.h"

#define PIN_MISO 12 //  SPI1 RX
#define PIN_CS 13 //    SPI1 CSn    CS
#define PIN_DC 7 //                 DC
#define PIN_RES 6 //                RES
#define PIN_SCK 14 //   SPI1 SCK    SCL
#define PIN_MOSI 15 //  SPI1 TX     SDA

#define SPI_PORT spi1
#define READ_BIT 0x80

int i, j, k;
//uint8_t R, G, B, pixel1, pixel2, R2, G2, B2, pixel3, pixel4;

static inline void cs_select() {
    asm volatile("nop \n nop \n nop");
    gpio_put(PIN_CS, 0);  // Active low
    asm volatile("nop \n nop \n nop");
}

static inline void cs_deselect() {
    asm volatile("nop \n nop \n nop");
    gpio_put(PIN_CS, 1);
    asm volatile("nop \n nop \n nop");
}

static void lxs_write_register(uint8_t reg) {
    cs_select();
    gpio_put(PIN_DC, 0);
    spi_write_blocking(SPI_PORT, &reg, 1);
    cs_deselect();
    //sleep_us(5);
}

static void write_data(uint8_t reg) {
    cs_select();
    gpio_put(PIN_DC, 1);
    spi_write_blocking(SPI_PORT, &reg, 1);
    cs_deselect();
    //sleep_us(500);
}

static void lxs_draw_line(uint8_t col_s, uint8_t row_s, uint8_t col_e, uint8_t row_e, uint8_t color_r, uint8_t color_g, uint8_t color_b) {
    uint8_t buf[8];
    buf[0] = 0x21;
    buf[1] = col_s;
    buf[2] = row_s;
    buf[3] = col_e;
    buf[4] = row_e;
    buf[5] = color_r;
    buf[6] = color_g;
    buf[7] = color_b;
    cs_select();
    gpio_put(PIN_DC, 0);
    spi_write_blocking(SPI_PORT, buf, 8);
    cs_deselect();
}

// "lxs_drawing_rectangle" not used
/*
static void lxs_drawing_rectangle(uint8_t col_s, uint8_t row_s, uint8_t col_e, uint8_t row_e, uint8_t line_color_r, uint8_t line_color_g, uint8_t line_color_b, uint8_t fill_color_r, uint8_t fill_color_g, uint8_t fill_color_b) {
    uint8_t buf[11];
    buf[0] = 0x22;
    buf[1] = col_s;
    buf[2] = row_s;
    buf[3] = col_e;
    buf[4] = row_e;
    buf[5] = line_color_r;
    buf[6] = line_color_g;
    buf[7] = line_color_b;
    buf[8] = fill_color_r;
    buf[9] = fill_color_g;
    buf[10] = fill_color_b;
    lxs_write_register(0x26); // Fill Enable / Disable
    lxs_write_register(1); // Enable Fill for Draw Rectangle
    cs_select();
    gpio_put(PIN_DC, 0);
    spi_write_blocking(SPI_PORT, buf, 11);
    cs_deselect();
    lxs_write_register(0x26); // Fill Enable / Disable
    lxs_write_register(0); // Disable Fill for Draw Rectangle
}
*/

static void lxs_clear_window(void) {
    uint8_t buf[5];
    buf[0] = 0x25;
    buf[1] = 0;
    buf[2] = 0;
    buf[3] = 95;
    buf[4] = 63;
    cs_select();
    gpio_put(PIN_DC, 0);
    spi_write_blocking(SPI_PORT, buf, 5);
    cs_deselect();
    sleep_ms(100);
}

static void lxs_write_numbers(float max_volt, uint8_t col, uint8_t row, uint8_t n) {
    int num;
    uint8_t col_start, col_end, row_start, row_end;
    uint8_t R, G, B, pixel1, pixel2, R2, G2, B2, pixel3, pixel4;
    char t[n];
    char buf[24];

    row_start = row;
    row_end = row + 6;

    snprintf(buf, 24, "%8.5f", max_volt);

    
    lxs_write_register(0x75); // Set Row Address
    lxs_write_register(row_start);
    lxs_write_register(row_end);
    //uint8_t font_1[9] = {0x1C, 0x36, 0x36, 0x1E, 0x06, 0x36, 0x1C, 0x00, 0x00};
    R = 31; G = 63; B = 31; //65k color : R (0-31), G (0-63), B (0-31)
    R2 = 0; G2 = 0; B2 = 0; //65k color : R (0-31), G (0-63), B (0-31)
    pixel1 = (R << 3) | (G >> 3);
    pixel2 = (G << 5) | B;
    pixel3 = (R2 << 3) | (G2 >> 3);
    pixel4 = (G2 << 5) | B2;
    for (k = 0; k < n; k++) {
        col_start = k * 7 + col;
        col_end = col_start + 4;
        lxs_write_register(0x15); // Set Column Address
        lxs_write_register(col_start);
        lxs_write_register(col_end);
        t[k] = buf[k] - 0x20;
        for (j = 0; j < 7; j++) {
            for(i = 0; i < 5; i++) {
                num = (Font[t[k]][i] >> j) & 1;
                if (num == 1) {
                    write_data(pixel1); // 8bits 65k format1 1st
                    write_data(pixel2); // 8bits 65k format1 2nd
                } else {
                    write_data(pixel3); // 8bits 65k format1 1st
                    write_data(pixel4); // 8bits 65k format1 2nd
                }
            }
        }
    }
}

static void lxs_write_string(char *sp, uint8_t col, uint8_t row, uint8_t n) {
    int num;
    uint8_t col_start, col_end, row_start, row_end;
    uint8_t R, G, B, pixel1, pixel2, R2, G2, B2, pixel3, pixel4;
    char t[n];

    row_start = row;
    row_end = row + 6;

    lxs_write_register(0x75); // Set Row Address
    lxs_write_register(row_start);
    lxs_write_register(row_end);
    R = 31; G = 63; B = 31; // 65k color : R (0-31), G (0-63), B (0-31)
    R2 = 0; G2 = 0; B2 = 0; // 65k color : R (0-31), G (0-63), B (0-31)
    pixel1 = (R << 3) | (G >> 3);
    pixel2 = (G << 5) | B;
    pixel3 = (R2 << 3) | (G2 >> 3);
    pixel4 = (G2 << 5) | B2;
    for (k = 0; k < n; k++) {
        col_start = k * 7 + col;
        col_end = col_start + 4;
        lxs_write_register(0x15); // Set Column Address
        lxs_write_register(col_start);
        lxs_write_register(col_end);
        t[k] = sp[k] - 0x20;
        for (j = 0; j < 7; j++) {
            for(i = 0; i < 5; i++) {
                num = (Font[t[k]][i] >> j) & 1;
                if (num == 1) {
                    write_data(pixel1); // 8bits 65k format1 1st
                    write_data(pixel2); // 8bits 65k format1 2nd
                } else {
                    write_data(pixel3); // 8bits 65k format1 1st
                    write_data(pixel4); // 8bits 65k format1 2nd
                }
            }
        }
    }
}

static void lxs_oled_init(void) {
    lxs_write_register(0xAE);
    lxs_write_register(0x25); // Clear Window
    lxs_write_register(0);
    lxs_write_register(0);
    lxs_write_register(95);
    lxs_write_register(63);
    lxs_write_register(0x15); // Set Column Address
    lxs_write_register(0);
    lxs_write_register(95);
    lxs_write_register(0x75); // Set Row Address
    lxs_write_register(0);
    lxs_write_register(63);
    lxs_write_register(0x81); // Set Contrast for Color "A" range:00d to 255d (128)
    lxs_write_register(128);
    lxs_write_register(0x82); // Set Contrast for Color "B" range:00d to 255d (128)
    lxs_write_register(128);
    lxs_write_register(0x83); // Set Contrast for Color "C" range:00d to 255d (128)
    lxs_write_register(128);
    lxs_write_register(0xA0); // Set Re-map & Data Format
    lxs_write_register(0b01110010); // 0x72 0b01110010
    lxs_write_register(0xA1);
    lxs_write_register(0x0);
    lxs_write_register(0xA2);
    lxs_write_register(0x0);
    lxs_write_register(0xA4);
    lxs_write_register(0xA8);
    lxs_write_register(0x3F);
    lxs_write_register(0xAD); // Set Master Configuration
    lxs_write_register(0x8E); // 0x8E
    lxs_write_register(0xB0);
    lxs_write_register(0x0B);
    lxs_write_register(0xB1);
    lxs_write_register(0x74);
    lxs_write_register(0xB3);
    lxs_write_register(0xD0);
    lxs_write_register(0x8A);
    lxs_write_register(0x80);
    lxs_write_register(0x8B);
    lxs_write_register(0x80);
    lxs_write_register(0x8C);
    lxs_write_register(0x80);
    lxs_write_register(0xBB);
    lxs_write_register(0x3E);
    lxs_write_register(0xBE);
    lxs_write_register(0x3E);
    lxs_write_register(0x87);
    lxs_write_register(0x0F);
    lxs_write_register(0xBE);
    lxs_write_register(0x3E);
    lxs_write_register(0xAF);
    sleep_ms(110);
}

void core1_entry() {

    const uint LED_PIN = 25;
    gpio_init(LED_PIN);
    gpio_set_dir(LED_PIN, GPIO_OUT);
    while (true) {
        gpio_put(LED_PIN, 1);
        sleep_ms(250);
        gpio_put(LED_PIN, 0);
        sleep_ms(250);
    }
}

int main() {
    int x1, roof = 26, central_line = 45, x_speed;
    float result_v, maximum_v, average_v, minimum_v;
    uint16_t y1, result;

    char strV[] = "V";
    char strMAX[] = "Max";
    char strAVE[] = "Ave";
    char strMIN[] = "Min";

    const float conversion_factor = 3.3f / (1 << 12);

    stdio_init_all();
    adc_init();

    multicore_launch_core1(core1_entry);

    // Make sure GPIO is high-impedance, no pullups etc
    adc_gpio_init(26);
    // Select ADC input 0 (GPIO26)
    adc_select_input(0);
    

    // This example will use SPI0 at 5MHz.
    spi_init(SPI_PORT, 5000 * 1000);
    gpio_set_function(PIN_MISO, GPIO_FUNC_SPI);
    gpio_set_function(PIN_SCK, GPIO_FUNC_SPI);
    gpio_set_function(PIN_MOSI, GPIO_FUNC_SPI);

    // Chip select is active-low, so we'll initialise it to a driven-high state
    gpio_init(PIN_CS);
    gpio_set_dir(PIN_CS, GPIO_OUT);
    gpio_put(PIN_CS, 1);

    gpio_init(PIN_RES);
    gpio_set_dir(PIN_RES, GPIO_OUT);
    gpio_put(PIN_RES, 1);
    gpio_init(PIN_DC);
    gpio_set_dir(PIN_DC, GPIO_OUT);
    
    lxs_oled_init();

    lxs_clear_window(); // It might not be necessary. (fuyou kamo)

    lxs_write_string(strMAX, 0, 0, 3); // saidai 8keta
    lxs_write_string(strV, 91, 0, 1); // saidai 8keta
    lxs_write_string(strAVE, 0, 9, 3); // saidai 8keta
    lxs_write_string(strV, 91, 9, 1); // saidai 8keta
    lxs_write_string(strMIN, 0, 18, 3); // saidai 8keta
    lxs_write_string(strV, 91, 18, 1); // saidai 8keta
    lxs_draw_line(0, central_line, 95, central_line, 200, 200, 200);

    uint16_t last_y1 = 63;
    uint8_t color_r = 255;
    uint8_t color_g = 255;
    uint8_t color_b = 255;
    x_speed = 920; // 1000

    while (1) {
        /*
        while (1) {
            result = adc_read();
            y1 = 57 - round(result * 15 * conversion_factor);
            if (y1 > central_line) {
                while (1) {
                    result = adc_read();
                    y1 = 57 - round(result * 15 * conversion_factor);
                    if (y1 < central_line) {
                        break;
                    }
                }
                break;
            }
        }
        */

        result_v = 0;
        maximum_v = 0;
        average_v = 0;
        minimum_v = 3.3;

        for(x1 = 0; x1 < 96; x1++){ // (x1 = 0; x1 < 96; x1++)
            result = adc_read();
            result_v = result * conversion_factor;
            if (maximum_v < result_v) {
                maximum_v = result_v;
            }
            if (minimum_v > result_v) {
                minimum_v = result_v;
            }
            average_v += result_v;
            y1 = 52 - round(result_v * 10); // Correct code -> y1 = 63 - round(result_v);
            if (x1 == 0) {
                last_y1 = y1;
            }
            if (y1 < central_line) {
                color_r = 0;
                color_g = 255;
                color_b = 0;
            } else {
                color_r = 255;
                color_g = 0;
                color_b = 0;
            }
            // Drawing a graph area
            lxs_draw_line(x1, roof, x1, 63, 0, 0, 0); // Delete old drawings (Fill in with black)
            lxs_draw_line(x1, central_line, x1, central_line, 200, 200, 200); // Drawing center line
            if (x1 > 0 & roof < y1 & roof < last_y1) { // Draw only specified range
                lxs_draw_line(x1 - 1, last_y1, x1, y1, color_r, color_g, color_b); // Draw Line
                last_y1 = y1;
            }
            sleep_us(x_speed); // 1000
        }
        // Numeric character drawing
        lxs_write_numbers(maximum_v, 30, 0, 8); // Maximum 8 digits 30, 0, 8
        lxs_write_numbers((average_v /= 96), 30, 9, 8); // Maximum 8 digits 30, 9, 8
        lxs_write_numbers(minimum_v, 30, 18, 8); // Maximum 8 digits 30, 18, 8
    }
}
 